#!/usr/bin/env python3

import subprocess
import numpy as np
import matplotlib.colors as clrs
import matplotlib.pyplot as plt
import matplotlib.ticker as tckr
from scipy.optimize import curve_fit


MATPLOTLIB_FIGSIZE = (2*6.4, 2*4.8)


#
# Load
#
# raw: kp, kd, tau
# kp: vertical y
# kd: horizontal x
#

raw = subprocess.check_output(
    ['cargo', 'run', '--', 'matrix', '5', '200', '5', '100'], cwd="pid-sim"
).decode('utf8')

size, *data = raw.strip().split('\n')
SIZE_KP, SIZE_KD = [int(f) for f in size.strip().split(';')]
parsed = [[float(f) for f in line.strip().split(';')] for line in data]
parsed = np.array(parsed).reshape((SIZE_KP, SIZE_KD, 3))

assert np.all(parsed[:,0,0] == parsed[:,1,0])
assert np.all(parsed[0,:,1] == parsed[1,:,1])


kps = parsed[:,0,0]
assert kps.shape == (SIZE_KP,)
kds = parsed[0,:,1]
assert kds.shape == (SIZE_KD,)
taus = parsed[:,:,2]
assert taus.shape == (SIZE_KP, SIZE_KD)


kp_mintaukds_idx = np.array([np.where(line == line.min())[0][0] for line in taus])
kp_mintaukds = kds[kp_mintaukds_idx]
assert kp_mintaukds_idx.shape == (SIZE_KP,)
assert kp_mintaukds.shape == (SIZE_KP,)

kd_mintaukps_idx = np.array([np.where(line == line.min())[0][0] for line in taus.T])
kd_mintaukps = kps[kd_mintaukps_idx]
assert kd_mintaukps_idx.shape == (SIZE_KD,)
assert kd_mintaukps.shape == (SIZE_KD,)

kp_mintaus = np.array([tausline[j] for (tausline, j) in zip(taus, kp_mintaukds_idx)])
kd_mintaus = np.array([tausline[j] for (tausline, j) in zip(taus.T, kd_mintaukps_idx)])


#
# Regresion (kp = f(tau) and kd = g(tau))
#

def regression_func(x, a, b):
    """
        x: tau
        a: k_scale
        b: k_exponent
    """
    return a * x**b

def fit_regression(x_data, y_data):
    popt, _pcov = curve_fit(regression_func, x_data, y_data, p0=(1, -1))
    return popt


kp_scale, kp_exponent = fit_regression(kp_mintaus, kps)
kps_reg = regression_func(kp_mintaus, kp_scale, kp_exponent)

cutoff_idx = np.where(kd_mintaus == kd_mintaus.min())[0][0]
kd_scale, kd_exponent = fit_regression(kd_mintaus[:cutoff_idx], kds[:cutoff_idx])
kds_reg = regression_func(kd_mintaus, kd_scale, kd_exponent)

print("kp_scale = %.6f" % kp_scale)
print("kp_exponent = %.6f" % kp_exponent)
print("kd_scale = %.6f" % kd_scale)
print("kd_exponent = %.6f" % kd_exponent)


#
# Show
#
# kp: horizontal x
# kd: vertical y
#

# 1

plt.figure(figsize=MATPLOTLIB_FIGSIZE)

plt.xlabel("Kp")
plt.ylabel("Kd")

plt.imshow(
    taus.T,
    aspect='auto',
    extent=[kps.min()-0.050, kps.max()+0.050, kds.min()-0.050, kds.max()+0.050],
    norm=clrs.LogNorm(),  # colorbar legend
)
cb = plt.colorbar(label="tau (sec)")
cb.ax.yaxis.set_major_formatter(tckr.FormatStrFormatter('%.1f'))

plt.plot(kps, kp_mintaukds, color='r')
#plt.plot(kd_mintaukps, kds, 'x', color='r')

plt.tight_layout()
plt.savefig("2-tau.png")

# 2

plt.figure(figsize=MATPLOTLIB_FIGSIZE)

plt.xlim((0, 4))
plt.xlabel("tau")

plt.plot(kp_mintaus, kps, 'x-')
plt.plot(kp_mintaus, kps_reg)

plt.plot(kd_mintaus, kds, 'x-')
plt.plot(kd_mintaus, kds_reg)

plt.legend(['Kp real', 'Kp regression', 'Kd real', 'Kd regression'])

plt.axhline(0, color="gray")

plt.tight_layout()
plt.savefig("3-regression.png")

# show

plt.show()
