use std::env;
use std::f64::consts::PI;
use std::process::exit;


fn usage(reason: &str) {
    println!("Error: {}", reason);
    println!("Usage:");
    println!("* <exe> one <kp> <kd>");
    println!("* <exe> one 1.000 -1.500");
    println!("* <exe> matrix <kp_start> <kp_end> <kd_start> <kd_end>");
    println!("* <exe> matrix 5 200 5 100");
    exit(1);
}


fn deg2rad(deg: f64) -> f64 {
    deg*PI/180.0
}


/// Simulates the system and its PID with the given weights
/// Returns the time constant tau (how fast the system reaches the setpoint)
fn engine_gimbal_pid(kp: f64, kd: f64, verbose: bool) -> f64 {
    let setpoint = deg2rad(2.0);                    // usually <4 deg for nominal flights
    let dt = 0.100;

    let mut tau = 0.0;                              // time constant to be returned
    let (mut ang_vel, mut ang_pos) = (0.0, 0.0);    // internal system variables

    for _t in 0..1000 {
        // yes, Rust is a bit stupid when it comes to for loops with floating point steps
        let t = _t as f64 * dt;

        // pid
        let err = setpoint - ang_pos;
        let derr = ang_vel;
        let control = kp*err + kd*derr;

        let mut ang_acc = control;

        // limit ang acc
        // if ang_acc.abs() > deg2rad(5.0) {
        //     ang_acc *= deg2rad(5.0)/ang_acc.abs();
        // }

        // sim
        ang_vel += ang_acc*dt;
        ang_pos += ang_vel*dt;

        // print info and check tau
        if verbose {
            println!("{:.6};{:.6};{:.6};{:.6}", t, ang_pos, ang_vel, ang_acc);
        }
        if err.abs() > 0.10*setpoint {
            tau = t;
        }
        assert!(ang_acc.abs() < deg2rad(50.0));
    }

    tau
}


fn main() {
    let args: Vec<String> = env::args().collect();

    if args.len() <= 1 {
        usage("Expects at least 2 args");
    }

    match args[1].as_str() {
        "one" => {
            if args.len() != 4 {
                usage("subcommand \"one\" expects 2 args");
            }

            let kp: f64 = args[2].parse::<f64>().unwrap();
            let kd: f64 = args[3].parse::<f64>().unwrap();

            let tau = engine_gimbal_pid(kp, kd, true);

            println!("{:.6};{:.6};{:.6}", kp, kd, tau);  // ~3
        },
        "matrix" => {
            if args.len() != 6 {
                usage("subcommand \"matrix\" expects 4 args");
            }

            let kp_start: i64 = args[2].parse::<i64>().unwrap();
            let kp_end: i64 = args[3].parse::<i64>().unwrap();
            let kd_start: i64 = args[4].parse::<i64>().unwrap();
            let kd_end: i64 = args[5].parse::<i64>().unwrap();

            println!("{};{}", kp_end-kp_start, kd_end-kd_start);

            for _kp in kp_start..kp_end {
                let kp = _kp as f64 * 0.100;

                for _kd in kd_start..kd_end {
                    let kd = - (_kd as f64) * 0.100;

                    let tau = engine_gimbal_pid(kp, kd, false);
                    println!("{:.6};{:.6};{:.6}", kp, kd, tau);
                }
            }
        },
        _ => {
            usage("Unexpected subcommand");
        },
    }
}
